
/* Portable Object Compiler */
/* Copyright (c) 1997,98 by Stes & Lerman.  All Rights Reserved. */

#include "objc1.h"
#include <string.h> /* strtok */
#include <stdlib.h> /* getenv */

/*
 * Most important stuff first.
 */

static void exitVersion(void)
{
	printf("1.8.2\n");
	exit(0);
}

static void printCopyright(void)
{
	STR s1,s2;
	s1 = "Portable Object Compiler 1.8.2\n";
	s2 = "Copyright (c) 1997,98 by Stes & Lerman.  All Rights Reserved.\n";
	printf(s1);
	printf(s2);
}

/*
 * Functions to parse command line arguments.
 */

static void unknownOption(id option)
{
	STR name = ([options cplus])?"objcpls1":"objc1";
	fprintf(stderr,"%s: Illegal option %s\n",name,[option str]);
	exit(1);
}

static void illegalArgument(id option,id arg)
{
	char *msg = "%s: Illegal argument %s for %s\n";
	STR name = ([options cplus])?"objcpls1":"objc1";
	fprintf(stderr,msg,name,[arg str],[option str]);
	exit(1);
}

/*
 * Use 'dash' character to distinguish between options and filenames.
 */

static BOOL isOption(id s)
{
	return [s size] && [s charAt:0] == '-';
}

/* reading options from string or file */

static id optionsFromString(id aCltn,char *s)
{
	char *p;
	char *delims = " \t\n\r";
	id buffer = [String str:s];
	
	p = strtok([buffer str],delims);
	
	while (p != NULL) {
		[aCltn add:[String str:p]];
		p = strtok(NULL,delims);
	}

	return aCltn;
}

static id optionsFromFile(id aCltn,FILE *f)
{
    char buf[BUFSIZ+1];

    while (!feof(f)) {
	if (fgets(buf,BUFSIZ,f)) {
	    /* this works 'cause strtok also deletes \n */
	    optionsFromString(aCltn,buf); 
	}
    }

    return aCltn;
}

/* this is mostly for MS-DOS cmd line length limitations */

static id optionsNamed(id aCltn,STR s)
{
	STR t;
	FILE *f;

	/* maybe there's such an environment variable */
	if ((t = getenv(s))) return optionsFromString(aCltn,t);

	/* maybe there's such a file */
	if ((f = fopen(s,"r"))) {
	    aCltn = optionsFromFile(aCltn,f);
	    fclose(f);return aCltn;
	}

	/* nope */
	return nil;	
}

/*
 * Cmd Line Options
 */

static id cmdLineOptions(int argc,char *argv[])
{
    int i;
    id aCltn = [Collection new];
    for(i=1;i<argc;i++) [aCltn add:[String str:argv[i]]];
    return aCltn;
}

/*
 * Filename for first line tag output (source filename)
 * (which we want to be file.m instead of file.P)
 */

static int setFileName(id option,id args)
{
	id arg = [args next];
	if (isOption(arg)) illegalArgument(option,arg);
	[options setSrcFileName:arg];
	return 0;
}

/*
 * Make initialization function customazible.
 */

static int setInitCall(id option,id args)
{
        id arg = [args next];
	if (isOption(arg)) illegalArgument(option,arg);
	[options setInitCall:arg];
	return 0;
}

static int setMainFunction(id option,id args)
{
        id arg = [args next];
	if (isOption(arg)) illegalArgument(option,arg);
	[options setMainFunction:arg];
	return 0;
}

static int setBind(id option,id args)
{
        id arg = [args next];
	if (isOption(arg)) illegalArgument(option,arg);
	[options setBind:arg];
	return 0;
}

/*
 * Make line max settable (works only for identifiers now).
 */

static int setLineMax(id option,id args)
{
	id arg = [args next];
	if (isOption(arg)) illegalArgument(option,arg);
	[options setLineMax:[arg asInt]];
	return 0;
}

/*
 * @ filename is understood to take options from "filename"
 * Maybe we should also allow @filename (without space)
 */

static void setOptions(id aCltn);

static int setOptionsFromFile(id option,id args)
{
        id arg = [args next];
	if (isOption(arg)) illegalArgument(option,arg);
	setOptions(optionsNamed([Collection new],[arg str]));
	return 0;
}

/*
 * Builtin functions (a la __builtin_isfloat).
 */

static int addBuiltinType(id option,id args)
{
	id arg = [args next];
	if (isOption(arg)) illegalArgument(option,arg);
	[[[Identifier str:[arg str]] markNoTypeDeclarator:YES] defineTypeName];
	return 0;
}

static int addBuiltinFun(id option,id args)
{
	id arg = [args next];
	if (isOption(arg)) illegalArgument(option,arg);
	[[Identifier str:[arg str]] defineBuiltinFunction];
	return 0;
}

/*
 * ANSI C book says #line is the directive.
 * GNU cc also accepts # lineno
 */

static void setShortTagFormat(void)
{
	id arg = [String str:"# %d \"%s\""];
	[options setLineTagFormat:arg];
}

/*
 * Function to parse command line arguments.
 */

static void setOptions(id aCltn)
{
	id args,s;
	int fileCount = 0; /* optional inFile and outFile */
	BOOL checkOption = YES; /* YES if filename can't begin with a dash */

        args = [aCltn eachElement];

	while ((s = [args next])) {
		if([s isEqualSTR:"-quiet"]) {[options setQuiet:YES];}
		else if([s isEqualSTR:"-z"]) {[options setOutputCode:NO];}
		else if([s isEqualSTR:"@"]){setOptionsFromFile(options,args);}
		else if([s isEqualSTR:"-u"]) {[options setBuffered:NO];}
		else if([s isEqualSTR:"-version"]) {[options setVersion:YES];}
		else if([s isEqualSTR:"-gnu"]){[options setGNU:YES];}
		else if([s isEqualSTR:"-noasm"]){[options setAsmKeyword:NO];}
		else if([s isEqualSTR:"-nolonglong"]){[options setLLKeyword:NO];}
		else if([s isEqualSTR:"-msdos"]) {[options setMSDOS:YES];}
		else if([s isEqualSTR:"-watcom"]) {[options setWATCOM:YES];}
#if __PORTABLE_OBJC__
		else if([s isEqualSTR:"-trace"]) {setMsgFlag(YES);}
#endif /* __PORTABLE_OBJC__ */
		else if([s isEqualSTR:"-comments"]){[options setComments:YES];}
		else if([s isEqualSTR:"-ibmvac"]){[options setIBMVAC:YES];}
		else if([s isEqualSTR:"-noSelTbl"]){
		    [options setDoSelTransTbl:NO];}
		else if([s isEqualSTR:"-noCategories"]){[options setCats:NO];}
		else if([s isEqualSTR:"-noBlocks"]){[options setBlocks:NO];}
		else if([s isEqualSTR:"-noCache"]){[options setCache:NO];}
		else if([s isEqualSTR:"-noFiler"]){[options setFilerCode:NO];}
		else if([s isEqualSTR:"-noSelfAssign"]){
		    [options setSelfAssign:NO];}
		else if([s isEqualSTR:"-noNilRcvr"]){[options setNilRcvr:NO];}
		else if([s isEqualSTR:"-objc"]){[options setTranslateCode:NO];}
		else if([s isEqualSTR:"-cplus"]){[options setCplus:YES];}
		else if([s isEqualSTR:"-boehm"]){[options setBoehm:YES];}
		else if([s isEqualSTR:"-cthreads"]){[options setCthreads:YES];}
		else if([s isEqualSTR:"-pthreads"]){[options setPthreads:YES];}
		else if([s isEqualSTR:"-inlinecache"]){
		    [options setInlineCache:YES];}
		else if([s isEqualSTR:"-refBind"]){[options setRefBind:YES];}
		else if([s isEqualSTR:"-export"]){setBind(s,args);}
		else if([s isEqualSTR:"-dllexport"]){
		    [options setBind:[String str:"__declspec(dllexport)"]];}
		else if([s isEqualSTR:"-traditional"]) {/* ignore */}
		else if([s isEqualSTR:"-filename"]){setFileName(s,args);}
		else if([s isEqualSTR:"-longlinetag"]) { /* compatib. */ }
		else if([s isEqualSTR:"-shortTags"]) {setShortTagFormat();}
		else if([s isEqualSTR:"-init"]) {setInitCall(s,args);}
		else if([s isEqualSTR:"-noShared"]) {
		    [options setSharedData:NO];}
		else if([s isEqualSTR:"-main"]) {setMainFunction(s,args);}
		else if([s isEqualSTR:"-linemax"]) {setLineMax(s,args);}
		else if([s isEqualSTR:"-builtinfunction"]) {
		    addBuiltinFun(s,args);}
		else if([s isEqualSTR:"-builtintype"]) {
		    addBuiltinType(s,args);}
		else if([s isEqualSTR:"-nolinetags"]) {
		    [options setLineTags:NO];}
		else if([s isEqualSTR:"-oneperfile"]) {
		    [options setOnePerFile:YES];}
		else if([s isEqualSTR:"-wClassUsedAsType"]) {
		    [options setWarnClassUsedAsType:NO];}
		else if([s isEqualSTR:"-wLocalInstance"]) {
		    [options setWarnLocalInstance:NO];}
		else if([s isEqualSTR:"-wUndefinedMethod"]) {
		    [options setWarnUndefinedMethod:NO];}
		else if([s isEqualSTR:"-wInterfaceNotFound"]) {
		    [options setWarnInterfaceNotFound:NO];}
		else if([s isEqualSTR:"-WLex"]) {[options setWarnLex:YES];}
		else if([s isEqualSTR:"-postlink"]) {[options setPostLink:YES];}
		else if([s isEqualSTR:"-checkbindfunction"]) {
		    [options setCheckBindCalled:YES];}
		else if([s isEqualSTR:"-debugInfo"]) {
		    [options setDebugInfo:YES];}
		else if([s isEqualSTR:"-ppi"]) {[options setPPI:YES];}
		else if([s isEqualSTR:"-"]) { checkOption = NO; }
		else if(checkOption && isOption(s)) {unknownOption(s);}
		else if(fileCount==0) {[options setInFile:s];fileCount++;}
		else if(fileCount==1) {[options setOutFile:s];fileCount++;}
		else {unknownOption(s);}
	}
}

/*
 * Opening and closing input file.  We're not dealing with wrapping around
 * files (include)  (because we're processing cpp output)
 */

extern FILE *yyin; /* scan.l global */

static void openInFile(id inFile)
{
	if (inFile) {
		yyin = openFile([inFile str],"r");
	} else {
		/* stdin */
	}

	/* initialize LineTag mechanism */
	/* everytime a newline is encountered, the scanner */
	/* will up the current LineTag object */

	[LineTag new];
}

static void closeInFile(id inFile)
{
	if (inFile) {
		fclose(yyin);
	} else {
		/* stdin */
	}
}

/*
 * Output code.
 */

static id outStream;

static void openOutFile()
{
	id outFile = [options outFile];

	if (outFile) {
		outStream = [OutputStream str:[outFile str]];
	} else {
		outStream = [OutputStream new]; /* stdout */
	}

	/* output a first LineTag */
	/* initialize output mechanism for LineTags */
	[LineTag outputOn:outStream];
}

static void closeOutFile()
{
	[LineTag finishOutputOn:outStream];
	[outStream close];
}

/*
 * After each declaration (top level definition), processNode()
 * is called to do semantical analysis (defining types etc.)
 * and then outputNode() to generate some code (if anything).
 */

void processNode(id aNode)
{
	if (aNode) [aNode process];
}

void outputNode(id aNode)
{
	BOOL doOutput = [options outputCode];
	BOOL doTranslate = [options translateCode];

	if (doOutput && aNode != nil) {
		if (doTranslate) {
		
			/* if node contains any blocks, output block definitions first */
			if ([options blocks]) {
			    [aNode processBlocks];
			    [aNode outputBlockDefsOn:outStream];
			}
		
			/* do code translation objc -> c */
			
			[aNode translateOn:outStream];
		} else {
			/* just dump objc to outstream */
			[aNode outputOn:outStream];
		}
	}
}


/*
 * Main entrance point of program.
 */

extern int yyparse(); /* parse.y entry point */

void main(int argc, char *argv[])
{
	[Options new]; /* options is a global */

	setOptions(cmdLineOptions(argc,argv));

	if ([options version]) {exitVersion();}
	if (![options quiet]) {printCopyright();}

	openOutFile();
	openInFile([options inFile]);

	[Module outputPrologueOn:outStream];
	if (yyparse()) exit(1);
	[Module outputEpilogueOn:outStream];

	closeInFile([options inFile]);
	closeOutFile();

	exit(exitStatus);
}
