
/*
 * Portable Object Compiler (c) 97,98 by Stes & Lerman.  All Rights Reserved.
 */

#include "Block.h"
#include <setjmp.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#define SPRINTF_BUFFER_SIZE (2048)

@implementation Block

/*****************************************************************************
 *
 * Creating Blocks
 *
 ****************************************************************************/

id OBJCRTEXPORT newBlock(int n,IMP f,void*d) 
{ 
    return [Block new:n:f:d]; 
}

/*
 * Might be nice to have a panel here, on the Macintosh
 * (otherwise the app just disappears and you have to look for 
 * a file "stderr" which contains (somewhere) the error message
 *
 */

static id err_fun(id thisBlock,void *data) 
{
    fprintf(stderr,"%s\n",(STR)data);
    return nil;
}

id OBJCRTEXPORT errorBlock(int tag,STR format,...) 
{ 
    STR msg;
    /* use OC macros for porting to SunOS4 */
    OC_VA_LIST ap;
    char aBuffer[SPRINTF_BUFFER_SIZE];

    OC_VA_START(ap,format);
    vsprintf(aBuffer,format,ap);
    OC_VA_END(ap);

    msg = OC_MallocAtomic(strlen(aBuffer) + 1);strcpy(msg,aBuffer);
    return [newBlock(0,(IMP)err_fun,msg) setTag:tag]; 
}

/*
 * Auxiliary function to implement 'catch' methods like
 * ifOutOfMemory: without using Block syntax
 *
 */

static id catch_fun(id thisBlock,void *data,id defaultHandler) 
{
    if ([defaultHandler tag] == [thisBlock tag]) {
	return [ (id) data valueOptional: defaultHandler ];
    } else {
	return [ defaultHandler raise ];
    }
}

static id catchBlock(int tag, id handler)
{
    return [newBlock(1,(IMP)catch_fun,handler) setTag:tag]; 
}

id idTrue;
id idFalse;

+ initialize
{
    idTrue = [Object new];
    idFalse = [Object new];
    return self;
}

- setUp:(int)n:(IMP)f:(void*)d
{
    nVars = n;fn = f;data = d;return self;
}

+ new:(int)n:(IMP)f:(void*)d
{
	return [[super new] setUp:n:f:d];
}

/*****************************************************************************
 *
 * Tags 
 *
 ****************************************************************************/

- (int) tag
{
    return tag;
}

- setTag:(int)aTag
{
    tag = aTag; return self;
}

/*****************************************************************************
 *
 * Exception Handling
 *
 ****************************************************************************/

#define MAX_JMP 32

static int jmpCount = 0;
static id jmpHandlers[MAX_JMP];
static jmp_buf jmpBuffers[MAX_JMP];

- raise
{
    if (jmpCount) {
	int thisCount = --jmpCount;

	/* perform handler as subroutine of raise */
	[jmpHandlers[jmpCount] valueOptional:self];

	/* unwind stack - jump to first handler that doesn't reraise */
	if (thisCount == jmpCount) longjmp(jmpBuffers[jmpCount],1);

	return nil; /* case handler did a resume */ 
    } else {
	[self value]; /* this might -error: and already abort */
	abort();
	return self;
    }
}

- ifException:aHandler
{
    jmpHandlers[jmpCount] = [aHandler goodHandler];
    
    if (setjmp(jmpBuffers[jmpCount])) {
	    return nil; /* indicates an exception was caught */
    } else {
	    jmpCount++;
	    if (jmpCount==MAX_JMP) return [self error:"increase MAX_JMP"];
	    return [self value];
    }
}

- resume
{
    /* prevent |raise| from doing a longjmp */
    jmpCount++;
    /* has the effect of pushing the exception handler back on the stack */
    return self;
}

- ifOutOfMemory:aBlock
{
    return [self ifException: catchBlock(OC_MEMERROR,aBlock)];
}

- ifClassNotFound:aBlock
{
    return [self ifException: catchBlock(OC_CLASSNOTFOUND,aBlock)];
}
    

/*****************************************************************************
 *
 * Evaluating Blocks
 *
 ****************************************************************************/

static char argMsg[] = "Block has wrong number of arguments";

- value
{
	return (nVars==0)?(id)(*fn)(self,data):[self error:argMsg];
}

static int atExitCount = 0;
static id atExitBlocks[32];

static void valueAtExit(void)
{
    if (atExitCount) [atExitBlocks[--atExitCount] value];
}

- atExit
{
    if (nVars == 0) {
	atExitBlocks[atExitCount++] = self;
	atexit(valueAtExit);
	return self;
    } else {
	return [self error:argMsg];
    }
}

- value:anObject
{
	return (nVars==1)?(id)(*fn)(self,data,anObject):[self error:argMsg];
}

- value:firstObject with:secondObject
{
	return (nVars==2)?(id)(*fn)(self,data,firstObject,secondObject):[self error:argMsg];
}

- valueOptional:anObject
{
    if (nVars == 1) return (id)(*fn)(self,data,anObject);
    if (nVars == 0) return (id)(*fn)(self,data);
    return [self error:argMsg];
}

- goodHandler
{
    return (nVars <= 1)?self:[self error:argMsg];
} 


/*****************************************************************************
 *
 * Control Flow 
 *
 ****************************************************************************/

+ idTrue
{
    return idTrue;
}

+ idFalse
{
    return idFalse;
}

- repeatTimes:(int)n
{
    int i;
    for(i=0;i<n;i++) [self value];
    return self;
}

- (BOOL) isFalse
{
    return NO; /* dummy for compilation with other compiler */
}

- (BOOL) isTrue
{
    return NO; /* dummy for compilation with other compiler */
}

- whileTrue:aBlock
{
    while ([[self value] isTrue]) [aBlock value];
    return self;
}

- whileFalse:aBlock
{
    while ([[self value] isFalse]) [aBlock value];
    return self;
}

- untilFalse:aBlock
{
    do [aBlock value]; while (![[self value] isFalse]);
    return self;
}

- untilTrue:aBlock
{
    do [aBlock value]; while (![[self value] isTrue]);
    return self;
}

@end

/* some defs needed when cross-compiling with a DIFFERENT compiler */
/* placed this here 'cause we will have a different objcrt.h and Object.h */

#ifndef __PORTABLE_OBJC__

void *OC_Malloc(size_t n) { return malloc(n); }
void *OC_MallocAtomic(size_t n) { return malloc(n); }
void *OC_Realloc(void *p,size_t n) { return realloc(p,n); }
void *OC_Free(void *p) { free(p);return NULL; }

#endif /* __PORTABLE_OBJC__ */

