
/*
 * Portable Object Compiler (c) 1997 by Stes & Lerman  All Rights Reserved.
 */

#include "oclib.h"

/*  class implementation (or interface) currently being parsed */
static id currentImplementation;

static id currentSetOfITypes; 
static id currentSetOfIVars; 
static id currentSetOfCTypes;
static id currentSetOfCVars;
static id currentIvarNodes;
static id currentCvarNodes;

static id collectSetOfTypes; /* for parsing */
static id collectSetOfVars;

static id currentInterface; /* for method translation */

static id setOfInterfaces;

@implementation ClassDefinition
- free
{
    return nil;
}


+ findInterface:aName
{
	if (!setOfInterfaces) setOfInterfaces = [Dictionary new];
	return [setOfInterfaces atKey:aName];
}

- addInterface
{
	assert(setOfInterfaces);
	[setOfInterfaces atKey:name put:self];
	return self;
}


- idName
{
	return name;
}

static void warnMultiple(id aName)
{
   char b[8192];
   char *s = [aName str];
   sprintf(b,"Found multiple interfaces for \"%s\"",s);
   warnError(b);
}

static void warnInterfaceNotFound(id aName)
{
   char b[8192];
   char *s = [aName str];
   sprintf(b,"Cannot find interface declaration for \"%s\"",s);
   warnError(b);
}

- (BOOL) isImplementationToken
{
    return [[subnodes at:0] isEqualSTR:"@implementation"];
}

- (BOOL) isInterfaceToken
{
    return [[subnodes at:0] isEqualSTR:"@interface"];
}

- processName:myName superName:hisName
{
	BOOL intfToken = [self isInterfaceToken];
	BOOL implToken = [self isImplementationToken];

	/* get string object from identifier */
	/* myName can be CLASSNAME_TOKEN or simply IDENTIFIER_TOKEN */
	/* (depending on whether it was defined before or not) */

	name = [myName string];

	/* interface is nil if class not yet defined */
	interface = [ClassDefinition findInterface:name];

	if (interface) {
	        if (intfToken) {
		    warnMultiple(myName);
		} else {
		    isImplementation = YES;
		    interfaceEmitted = YES;
		}
	} else {
		/* this is the case of either an interface declaration
		 * or the case of an implementation that is at the same
		 * time its own interface
		 */
	        
		if (!intfToken) [self setIsImplementation:YES];

	        /* issue a warning, but not for '=' class defs */
		if (implToken && [options warnInterfaceNotFound]) {
		    warnInterfaceNotFound(myName);
		}
	
		/* make the identifier a special token (CLASSNAME_TOKEN) */
		[myName defineClassName];

		/* declare class interface, set interface to self */
		interface = [self addInterface];
	}

	if (myName != nil && hisName != nil && [hisName isEqual:myName]) {
	    fatalError("Name of class and superclass equal");
	    return self;
	}

	/* get string object from identifier */
	superName = (hisName)?[hisName string]:nil;
	
	if (interface != self) {
	    /* interface superName can be nil for rootclass */
	    if ([interface superName]) {
		if (superName) {
		   if (![superName isEqual:[interface superName]])
			fatalError("Found conflicting superclass");
		}
	    } else {
		if (superName) 
			fatalError("Found conflicting superclass");
	    }
	}

	return self;
}

- superName
{
	return superName;
}

- (BOOL) isCategory
{
        return isCategory;
}

- setIsCategory:(BOOL)flag
{
        isCategory = flag; return self;
}

/* Very primitive mechanism to collect the instance variables.
 * We need them to expand something like "ivar" to "self->ivar",
 * but at this moment, just remember the name, and a node
 * representation of the type.
 */

+ declareIvar:aNode Itype:bNode
{
   if (!collectSetOfVars) collectSetOfVars = [Collection new];
   if (!collectSetOfTypes) collectSetOfTypes = [Collection new];
   [aNode addIvarsTo:collectSetOfVars addItype:bNode to:collectSetOfTypes];
   return self;
}

- completeSetOfIvars
{
	if (completeSetOfIvars) {
		return completeSetOfIvars;
	} else {
		id face;
	/* not a Collection - idea is to have fast lookup (contains:) */
		completeSetOfIvars = [Set new];
		for(face = self;face;face = [face superInterface]) {
			[completeSetOfIvars addContentsOf:[face setOfIVars]];
		}
		return completeSetOfIvars;
	}
}

- completeSetOfCvars
{
	if (completeSetOfCvars) {
		return completeSetOfCvars;
	} else {
		id face;
	/* not a Collection - idea is to have fast lookup (contains:) */
		completeSetOfCvars = [Set new];
		for(face = self;face;face = [face superInterface]) {
			[completeSetOfCvars addContentsOf:[face setOfCVars]];
		}
		return completeSetOfCvars;
	}
}

+ (BOOL) inMethodImplementation
{
    return (currentInterface != nil);
}

+ (BOOL) doesCurrentSetOfIVarsContain:anIdentifier
{
	if (currentInterface) {
	return [[currentInterface completeSetOfIvars] contains:anIdentifier];
	} else {
	/* outside method implementation, can still be in class impl */
	return NO;
	}
}

+ (BOOL) doesCurrentSetOfCVarsContain:anIdentifier
{
	if (currentInterface) {
	return [[currentInterface completeSetOfCvars] contains:anIdentifier];
	} else {
	/* outside method implementation, can still be in class impl */
	return NO;
	}
}

- processVars
{
	if (currentSetOfIVars) {
	        ivarNodes   = currentIvarNodes;
		setOfIVars  = currentSetOfIVars;
		setOfITypes = currentSetOfITypes;
		assert([setOfIVars size] == [setOfITypes size]);
		currentIvarNodes   = nil;
		currentSetOfIVars  = nil;
		currentSetOfITypes = nil;

	} else {
		/* no new ivars in this subclass */
		ivarNodes   = nil;
		setOfIVars  = [Collection new];
		setOfITypes = [Collection new];
	}

	if (currentSetOfCVars) {
	        cvarNodes   = currentCvarNodes;
		setOfCVars  = currentSetOfCVars;
		setOfCTypes = currentSetOfCTypes;
		assert([setOfCVars size] == [setOfCTypes size]);
		currentCvarNodes   = nil;
		currentSetOfCVars  = nil;
		currentSetOfCTypes = nil;

	} else {
		/* no new cvars in this subclass */
		cvarNodes   = nil;
		setOfCVars  = [Collection new];
		setOfCTypes = [Collection new];
	}

	/* if duplicated in both @interface and @implementation */
	/* they have to literally match */

	if (interface != self && (currentSetOfCVars || currentSetOfIVars)) {
	      if (![ivarNodes isEqual:[interface ivarNodes]]) {
		   fatalError("conflicting instance variable declaration");
	      }
	      if (![cvarNodes isEqual:[interface cvarNodes]]) {
		   fatalError("conflicting class variable declaration");
	      }
	}

	return self;
}

- superInterface
{
	if (superName) {
		id superInterface = [ClassDefinition findInterface:superName];
		assert(superInterface);
		return superInterface;
	} else {
		return nil;
	}
}

- rootInterface
{
	id rootface;
	id superface;
	
	rootface = self;
	while ((superface = [rootface superInterface])) rootface = superface; 
	return rootface;
}

- rootName
{
	id rootface = [self rootInterface];
	assert(rootface);
	return [rootface idName];
}

+ collectIvarNodes:aNode
{
	currentIvarNodes = aNode; /* for literal dumping as in @defs */
	currentSetOfIVars = collectSetOfVars;
	currentSetOfITypes = collectSetOfTypes;
	collectSetOfVars = nil;
	collectSetOfTypes = nil;
	return self;
}

+ collectCvarNodes:aNode
{
	currentCvarNodes = aNode; /* for literal dumping as in @defs */
	currentSetOfCVars = collectSetOfVars;
	currentSetOfCTypes = collectSetOfTypes;
	collectSetOfVars = nil;
	collectSetOfTypes = nil;
	return self;
}

- makeCurrent
{
	currentInterface = self;
	return self;
}

- setOfIVars
{
	return setOfIVars;
}

- setOfCVars
{
	return setOfCVars;
}

- setOfIdIVars
{
	if (setOfIdIVars) {
		return setOfIdIVars;
	} else {
		int i,n;
		setOfIdIVars = [Collection new];
		for(i=0,n=[setOfIVars size];i<n;i++) {
			if ([[setOfITypes at:i] isIdType]) {
				[setOfIdIVars add:[setOfIVars at:i]];
			}
		}
		return setOfIdIVars;
	}
}

- ivarNodes
{
	return ivarNodes;
}

- cvarNodes
{
	return cvarNodes;
}

- clearCurrent
{
	currentInterface = nil;
	return self;
}

- outputIVarsCommentOn:aStream
{
	id ivar,ivars = [setOfIVars eachElement];
	[aStream printStr:"\n/*\n"];
	[aStream printStr:" * %s instance variables\n",[name str]];
	while ((ivar = [ivars next])) {
		[aStream printStr:" *"];
		[ivar outputWithoutTagsOn:aStream];
		[aStream printNewline];
	}
	[ivars free];
	[aStream printStr:" */\n"];
	return self;
}


- outputIvarNodesOn:aStream
{
	if (superName) [[self superInterface] outputIvarNodesOn:aStream];
	if (ivarNodes) [ivarNodes outputOn:aStream];
	
	return self;
}

- outputCvarNodesOn:aStream
{
	if (superName) [[self superInterface] outputCvarNodesOn:aStream];
	if (cvarNodes) [cvarNodes outputOn:aStream];
	
	return self;
}

- (STR) classGlobalFun
{
    return [[Identifier str:[name str]] classGlobalFun];
}

- (STR) classStructFun
{
    return [[Identifier str:[name str]] classStructFun];
}

- (STR) metaStructFun
{
    return [[Identifier str:[name str]] metaStructFun];
}

- outputExternClassGlobalOn:aStream
{
    STR e = [[options bind] str]; /* stuff like __export etc. */
    [aStream printNewline];
    if ([options sharedData]) {
	[aStream printStr:"extern id %s %s;\n",e,[name str]];
    } else {
	[aStream printStr:"extern id %s *%s(void);\n",e,[self classGlobalFun]];
    }
    return self;
}

- translateInterfaceOn:aStream
{
	[ self outputExternClassGlobalOn:aStream];

        /* definition of our struct _SHARED */
        /* stes sep/2/97 this used to be in Module's prologue */
        /* moved to this location for class variables */

	/* in fact we only need struct shared of the meta, the class
	 * being implemented and of its superclass, but just emitting
	 * it for every interface is easier (and avoids duplication)
	 */

        [self outputStructSharedTypeOn:aStream];
        [aStream printNewline];

	[LineTag forceLineTag];
	interfaceEmitted = YES;
	return self;
}


- setIsImplementation:(BOOL)flag
{
	isImplementation = flag;return self;
}

+ checkForPendingDefinitions
{
    if (currentImplementation) {
	    id aMsg = [currentImplementation idName];
	    aMsg = [String sprintf:"Class %s not properly ended.",[aMsg str]];
	    fatalError([aMsg str]);return nil;
    } else {
	    return self;
    }
}

+ new
{
	[self checkForPendingDefinitions];
	currentImplementation = [super new];
	return currentImplementation;
}

+ end
{	
	if (currentImplementation) {
	    
		id prev = currentImplementation;
		
	    /* set global back to nil to end implemenation context */
	    currentImplementation = nil;
		
	    return prev;
	} else {
	    fatalError("Class end outside class definition context");
	    return nil;
	}
}

+ currentName
{
	return [currentImplementation idName];	
}

+ current
{
	return currentImplementation;	
}


- addClsDispatchTable:aMethod
{
	if (!clsDispatchTable) clsDispatchTable = [Collection new];
	[clsDispatchTable add:aMethod];
	return self;
}

- addNstDispatchTable:aMethod
{
	if (!nstDispatchTable) nstDispatchTable = [Collection new];
	[nstDispatchTable add:aMethod];
	return self;
}

- doModuleDefinitions
{
	id sName;
	
	assert(isImplementation);

	if ([options onePerFile]) {
		[Module setTheOnlyClassImplementation:self];
	} else {
		[Module addClassImplementation:self];
	}

	sName = [interface superName]; /* can be nil */
	if (sName) [Module usesEntry:sName];

	/* add entries for this implementation to the Module */
	[Module definesEntry:name];
	
	return self;
}


/*
 * definition of the Objective C class struct
 * for compatibility with Stepstone's runtime
 * there are some fields that we don't use (attr & objcID)
 *
 * "isa" is struct _SHARED for Stepstone, but I like "id" better.
 *
 * stes apr29/97
 * Changed output of struct _PRIVATE to struct Class_PRIVATE.
 * You can then in a C source level debugger cast to the 
 * struct _PRIVATE of some class.
 */

- (STR) structPrivateName
{
    return [[String sprintf:"%s_PRIVATE",[name str]] str];
}

- outputStructPrivateTypeOn:aStream
{
[aStream printStr:"struct %s {",[self structPrivateName]];
/* instance variables */
[LineTag forceLineTag];
[interface outputIvarNodesOn:aStream];
[aStream printStr:"\n};\n"];

/* for debugging - this list should match the def above */
if ([options comments]) [interface outputIVarsCommentOn:aStream];

return self;
}

/*
 * stes sep/1/97
 * definition of "struct _SHARED" for a class
 * moved this from module.m to classdef.m because it is now possibly
 * different for each class (because of support for class variables)
 * 
 */

/*
 * definition of the Objective C class struct
 * for compatibility with Stepstone's runtime, our
 * extensions come after the "extensions" comment
 *
 * Note that the reference to struct _SLT is a pointer,
 * so we can do this without defining the type.
 * (objcc defines it in a header, and does NOT generate it).
 * So for cross-compiles, we have to do the same.
 */

- (STR) structSharedName
{
    return [[String sprintf:"%s_SHARED",[name str]] str];
}

#if 0
static char *sharedType = "\n"
	"struct _SHARED {\n"
/* keep this compatible with objcc */
	"\tstruct _SHARED  *isa;\n"
	"\tstruct _SHARED  *clsSuper;\n"
	"\tchar 		*clsName;\n"
	"\tchar		*clsTypes;\n"
	"\tshort 		clsSizInstance;\n"
	"\tshort 		clsSizDict;\n"
	"\tstruct _SLT 	*clsDispTable;\n"
	"\tlong 		clsStatus;\n"
	"\tstruct modDescriptor *clsMod;\n"
/* not all objcc 4.x have clsVersion */
	"\tunsigned	clsVersion;\n"
/* extensions */
	"\tstruct _SHARED       *clsCats;\n"
	"};\n"
;
#endif

- outputStructSharedTypeOn:aStream
{
	STR superShared;
        STR thisShared  = [interface structSharedName];
	STR metaShared  = [[interface rootInterface] structSharedName]; 

        /* superclass can be nil, type nil as metaShared */
        if ([interface superInterface]) {
	    superShared = [[interface superInterface] structSharedName];
	} else {
	    superShared = metaShared;
	}

        [aStream printStr:"struct %s {\n",thisShared];

        /* keep this compatible with objcc's struct _SHARED */

	[aStream printStr:"\tstruct %s *isa;\n",metaShared];
	[aStream printStr:"\tstruct %s *clsSuper;\n",superShared];
	[aStream printStr:"\tchar *clsName;\n"];
	[aStream printStr:"\tchar *clsTypes;\n"];
	[aStream printStr:"\tshort clsSizInstance;\n"];
	[aStream printStr:"\tshort clsSizDict;\n"];
	[aStream printStr:"\tstruct _SLT *clsDispTable;\n"];
	[aStream printStr:"\tlong clsStatus;\n"];
	[aStream printStr:"\tstruct modDescriptor *clsMod;\n"];

        /* not all objcc 4.x have clsVersion */

	[aStream printStr:"\tunsigned clsVersion;\n"];

        /* Portable Object Copmiler extensions begin here */

	[aStream printStr:"\tstruct %s *clsCats;\n",metaShared];
	[aStream printStr:"\tid *clsGlbl;\n"];
	
	/* class variables */

	[self outputCvarNodesOn:aStream];

	[aStream printStr:"\n};\n"];

        return self;
}

static void printSLT(id aStream,id aMethod,id className)
{
STR s = [[aMethod fullMethodName] str];
STR t = [[aMethod functionNameForClass:className] str];
[aStream printStr:"\n{\"%s\",(id (*)())%s},",s,t];
}

/*
* SGI cc doesn't like (aborts on) empty declarations
* so always output at least {0,0}
*/

static void printNullSLT(id aStream)
{
[aStream printStr:"\n{(char*)0,(id (*)())0}"];
}


- (STR) _clsDispatchTblName
{
	if ([options onePerFile]) {
		return "_clsDispatchTbl";
	} else {
		return [[String sprintf:"_%s_clsDispatchTbl",[name str]] str];
	}
}

- (STR) _nstDispatchTblName
{
	if ([options onePerFile]) {
		return "_nstDispatchTbl";
	} else {
		return [[String sprintf:"_%s_nstDispatchTbl",[name str]] str];
	}
}

- outputClsDispatchTablesOn:aStream
{
int i,n;

n = (clsDispatchTable)?[clsDispatchTable size]:0;

[aStream printNewline];
[aStream printStr:"static struct _SLT %s[] ={",[self _clsDispatchTblName]];

for (i=0;i<n;i++) printSLT(aStream,[clsDispatchTable at:i],[self idName]);
printNullSLT(aStream);
[aStream printStr:"\n};\n"];
[aStream printNewline];

return self;
}

- outputNstDispatchTablesOn:aStream
{
int i,n;

n = (nstDispatchTable)?[nstDispatchTable size]:0;

[aStream printNewline];
[aStream printStr:"static struct _SLT %s[] ={",[self _nstDispatchTblName]];
for (i=0;i<n;i++) printSLT(aStream,[nstDispatchTable at:i],[self idName]);
printNullSLT(aStream);
[aStream printStr:"\n};\n"];
[aStream printNewline];

return self;
}

- outputDispatchTablesOn:aStream
{
STR sName = [name str];

if ([options comments]) {
[aStream printNewline];
[aStream printStr:"\n/*\n"];
[aStream printStr:" * %s Dispatch tables \n",sName];
[aStream printStr:" */\n"];
[aStream printNewline];
}

[self outputClsDispatchTablesOn:aStream];
[self outputNstDispatchTablesOn:aStream];

return self;
}

/* clsStatus : these must match objcrt.m */
#define CLS_FACTORY 0x1L
#define CLS_META 0x2L
#define CLS_INITIALIZED 0x4L
#define CLS_POSING 0x8L
#define CLS_MAPPED 0x10L
#define CLS_CATS 0x20L /* obsolete, will go away */
#define CLS_CAT 0x40L 

static void comment(id aStream,STR comment)
{
	if ([options comments]) [aStream printStr:comment];
	[aStream printNewline];
}

- outputStructSharedClassOn:aStream
{
STR strSuperName;
STR strName = [name str];

strSuperName=([interface superName])?[[interface superName] str]:NULL;	

if ([options comments]) {
	[aStream printNewline];
	[aStream printStr:"\n/*\n"];
	[aStream printStr:" * struct for class %s \n",strName];
	[aStream printStr:" */\n"];
	[aStream printNewline];
}

/* name of struct */
[aStream printStr:"struct %s ",[interface structSharedName]];

[aStream printStr:" _%s = {\n",strName];

/* isa of our isa = our meta */
[aStream printStr:"\t&__%s,",strName];comment(aStream,"/* isa */");

/* super of our isa = isa of our super */
/* in the -noShared case this field is set in BIND() */
if (strSuperName && [options sharedData]) {
	[aStream printStr:"\t&_%s,",strSuperName];
} else {
	[aStream printStr:"\t0,"]; /* ends super hierarchy */
}
comment(aStream,"/* clsSuper */");

/* clsName */
[aStream printStr:"\t\"%s\",",strName];
comment(aStream,"/* clsName */");

/* clsTypes : no filer types yet */
[aStream printStr:"\t0,"];
comment(aStream,"/* clsTypes */");

/* let C compiler compute size */
[aStream printStr:"\tsizeof(struct %s_PRIVATE),",strName];
comment(aStream,"/* clsSizInstance */");

/* clsSizDict */
[aStream printStr:"\t%d,",(nstDispatchTable)?[nstDispatchTable size]:0];
comment(aStream,"/* clsSizDict */");

/* _nstDispatchTbl */
[aStream printStr:"\t%s,",[self _nstDispatchTblName]];
comment(aStream,"/* _nstDispatchTbl */");

/* clsStatus */
if ([interface isCategory]) {
[aStream printStr:"\t%d,",CLS_FACTORY | CLS_POSING | CLS_CATS | CLS_CAT];
} else {
[aStream printStr:"\t%d,",CLS_FACTORY | CLS_POSING | CLS_CATS];
}
comment(aStream,"/* clsStatus */");

/* clsMod */
[aStream printStr:"\t&%s,",[[Module moduleDescName] str]];
comment(aStream,"/* clsMod */");

/* clsVersion */
[aStream printStr:"\t0,"];
comment(aStream,"/* clsVersion */");

/* clsCats - obsoleted */
[aStream printStr:"\t0,"];
comment(aStream,"/* clsCats */");

/* clsGlbl */
[aStream printStr:"\t&%s,",strName];
comment(aStream,"/* clsGlbl */");

/* end of class struct */
[aStream printStr:"};\n"];

/* output for -noShare and -share */
[aStream printStr:"struct %s %s *%s(void) { return &_%s; }\n",
	[interface structSharedName],[[options bind] str],
	[self classStructFun],strName];

return self;
}

- outputStructSharedMetaClassOn:aStream
{
STR strSuperName;
STR strName = [name str];
STR rootName = [[interface rootName] str];	
STR metaShared = [[interface rootInterface] structSharedName];

strSuperName=([interface superName])?[[interface superName] str]:NULL;	

assert(rootName);

if ([options comments]) {
[aStream printNewline];
[aStream printStr:"\n/*\n"];
[aStream printStr:" * %s MetaClass struct\n",strName];
[aStream printStr:" */\n"];
[aStream printNewline];
}

/* name of struct */
[aStream printStr:"struct %s ",metaShared];

[aStream printStr:" __%s = {\n",strName];

if ([options sharedData]) {
    /* isa of our meta = meta of Object */
    [aStream printStr:"\t&__%s,",rootName];
} else {
    [aStream printStr:"\t0,"]; /* field set in BIND() */
}
comment(aStream,"/* isa */");

if ([options sharedData]) {
    /* super of our meta = meta of our super */
    if (strSuperName) {
	[aStream printStr:"\t&__%s,",strSuperName];
    } else {
	/* stes jan25, 1997
	 * The meta hierarchy is ended NOT by a '0'
	 * but rather by making it point to the root class
	 * itself which has a super of '0'.  This makes classes
	 * respond to Object instance methods, such as |name|
	 * (This was done wrong for objc1.1.1)
	 * In other words: a search doesn't end with +name, but
	 * with -name.
	 */
	[aStream printStr:"\t&_%s,",rootName];
    }
} else {
    [aStream printStr:"\t0,"]; /* field set in BIND() */
}
comment(aStream,"/* clsSuper */");

/* clsName */
[aStream printStr:"\t\"%s\",",strName];
comment(aStream,"/* clsName */");

/* clsTypes : no filer types yet */
[aStream printStr:"\t0,"];
comment(aStream,"/* clsTypes */");

/* let C compiler compute size */
[aStream printStr:"\tsizeof(struct %s),",[interface structSharedName],strName];
comment(aStream,"/* clsSizInstance */");

/* clsSizDict */
[aStream printStr:"\t%d,",(clsDispatchTable)?[clsDispatchTable size]:0];
comment(aStream,"/* clsSizDict */");

/* _nstDispatchTbl */
[aStream printStr:"\t%s,",[self _clsDispatchTblName]];
comment(aStream,"/* _clsDispatchTbl */");

/* clsStatus */
[aStream printStr:"\t%d,",CLS_META | CLS_CATS];
comment(aStream,"/* clsStatus */");

/* clsMod */
[aStream printStr:"\t&%s,",[[Module moduleDescName] str]];
comment(aStream,"/* clsMod */");

/* clsVersion */
[aStream printStr:"\t0,"];
comment(aStream,"/* clsVersion */");

/* clsCats - obsoleted */
[aStream printStr:"\t0,"];
comment(aStream,"/* clsCats */");

/* clsGlbl */
[aStream printStr:"\t0,"];
comment(aStream,"/* clsGlbl */");

/* end of class struct */
[aStream printStr:"};\n"];
/* output for -noShare and -share */
[aStream printStr:"struct %s %s *%s(void) { return &__%s; }\n",
	metaShared,[[options bind] str],
	[self metaStructFun],strName];
return self;
}

- outputClassGlobalFunCallsOn:aStream
{
    STR t;
    STR s = [name str];
    id rootface = [interface rootInterface];
    id superface = [interface superInterface];

    /* this is called for each class in a module, inside OBJCBIND()
     * for the module.
     * allows us, in the -noShare case, to call a function and like
     * this fill out fields in the struct shared and meta of a class
     * (all necessary because problems with shared data with DLL's)
     */

    /* '0' ends super hierarchy, otherwise call function */
    if (superface) {
	t = [superface classStructFun];
	[aStream printStr:"\t_%s.clsSuper = %s();\n",s,t];
    }

    /* meta isa and clsSuper are never '0' */
    t = [rootface metaStructFun];
    [aStream printStr:"\t__%s.isa = %s();\n",s,t];

    /* '0' does NOT end meta hierarchy (see comment above) */
    if (superface) {
	t = [superface metaStructFun];
    } else {
        t = [rootface classStructFun];
    }
    [aStream printStr:"\t__%s.clsSuper = %s();\n",s,t];

    return self;
}

- translatePrologueOn:aStream
{
STR s = [name str];
STR t = ([interface superName])?[[interface superName] str]:NULL;	
STR r = [[interface rootName] str];	
STR superShared;
STR thisShared = [interface structSharedName];
STR metaShared = [[interface rootInterface] structSharedName];
STR e = [[options bind] str];

superShared = (t)?[[interface superInterface] structSharedName]:metaShared;

if ([options comments]) {
[aStream printNewline];
[aStream printStr:"\n/*\n"];
[aStream printStr:" * %s implementation\n",s];
[aStream printStr:" */\n"];
[aStream printNewline];
}

/* forward declaration of class             */
/* already happened by the Class interface  */
/* the static global in the -noShared case, best no in interface */

if ([options sharedData]==NO) [aStream printStr:"static id %s;\n",s]; 

/* forward declaration of meta of root class */
/* used for our struct _SHARED */
/* root class of root class is itself */

[aStream printNewline];
[aStream printStr:"extern struct %s __%s;\n",metaShared,r];
if ([options sharedData]==NO) {
    [aStream printStr:"extern struct %s %s * %s(void);\n",
	metaShared,e,[[interface rootInterface] metaStructFun]];
}

/* forward declaration of class and metaclass structs     */
/* for superClass.  these are used for our struct _SHARED */
/* not needed if we're defining a rootClass (would be NULL) */
if (t) {
    [aStream printStr:"extern struct %s _%s;\n",superShared,t];
    [aStream printStr:"extern struct %s __%s;\n",metaShared,t];
    if ([options sharedData]==NO) {
        [aStream printStr:"extern struct %s %s * %s(void);\n",
	   superShared,e,[[interface superInterface] classStructFun]];
        [aStream printStr:"extern struct %s %s * %s(void);\n",
	   metaShared,e,[[interface superInterface] metaStructFun]];
    }
}

/* forward declaration of class and metaclass structs */
/* these are referred to in messages to "super"       */
[aStream printStr:"extern struct %s _%s;\n",thisShared,s];
[aStream printStr:"extern struct %s __%s;\n",metaShared,s];
if ([options sharedData]==NO) {
    [aStream printStr:"extern struct %s %s * %s(void);\n",
       thisShared,e,[interface classStructFun]];
    [aStream printStr:"extern struct %s %s * %s(void);\n",
       metaShared,e,[interface metaStructFun]];
}

/* definition of "struct _PRIVATE" */
	[self outputStructPrivateTypeOn:aStream];
	[aStream printNewline];

 /* ended flag is needed in the context of implementations
  * for whether to emit a prologue or epilogue.
  * for interfaces it doesn't matter
  */
		
	prologueEmitted = YES;
	
	/* module defs for the references we have made in this method */
	[self doModuleDefinitions];

	[LineTag forceLineTag];
	return self;
}

/******************************************************************************
 *
 *           A S C I I F I L E R   M E T H O D S
 *
 *****************************************************************************/

- outputSuperMessage:aSel argument:(STR)argName on:aStream
{
	STR s = [name str];

	[aStream printStr:"\t(*(id (*) (id,SEL,id))_impSuper((id)__%s.clsSuper,",s];
	[aStream printSelector:aSel];
	[aStream printStr:"))((id)self,"];
	[aStream printSelector:aSel];
	[aStream printStr:",%s);\n",argName];

	return self;
}

- outputSuperFileOutMessageOn:aStream
{
	id aSel = [String str:"fileOutIdsFor:"];
	return [self outputSuperMessage:aSel argument:"aFiler" on:aStream];
}

- outputSuperFileInMessageOn:aStream
{
	id aSel = [String str:"fileInIdsFrom:"];
	return [self outputSuperMessage:aSel argument:"aFiler" on:aStream];
}

- outputFilerMessage:aSel ivar:ivar on:aStream
{
	[aStream printStr:"\t(*(id (*) (id,SEL,void*,char))_imp(aFiler,"];
	[aStream printSelector:aSel];
	[aStream printStr:"))(aFiler,"];
	[aStream printSelector:aSel];
	[aStream printStr:",&"];
	[ivar translateOn:aStream asIvarOf:name];
	[aStream printStr:",'@');\n"];
	return self;
}

- outputFileOutMessage:ivar on:aStream
{
	id aSel = [String str:"fileOut:type:"];
	return [self outputFilerMessage:aSel ivar:ivar on:aStream];
}

- outputFileInMessage:ivar on:aStream
{
	id aSel = [String str:"fileIn:type:"];
	return [self outputFilerMessage:aSel ivar:ivar on:aStream];
}

static id implementFileOutMethod(void)
{
	id method = [MethodInfo new];
	[method setIsFactory:NO];
	[method setResultType:[Token str:"id"]];
	[method addNameComponent:[String str:"fileOutIdsFor:"]];
	[method addArgumentType:[Token str:"id"]];
	[method addArgument:[Identifier str:"aFiler"]];
	
	/* add necessary info to class structures */
	[method implement];
	return method;
}

static id implementFileInMethod(void)
{
	id method = [MethodInfo new];
	[method setIsFactory:NO];
	[method setResultType:[Token str:"id"]];
	[method addNameComponent:[String str:"fileInIdsFrom:"]];
	[method addArgumentType:[Token str:"id"]];
	[method addArgument:[Identifier str:"aFiler"]];

	/* add necessary info to class structures */
	[method implement];
	return method;
}

- outputFileOutCodeOn:aStream
{
	id ivars,ivar,method = implementFileOutMethod();
	
	[method outputFunctionPrototypeOn:aStream];
	[aStream printStr:"\n{\n"];

	/* generate call to super */
	if ([interface superName]) {
		[self outputSuperFileOutMessageOn:aStream];
		ivars = [[interface setOfIdIVars] eachElement];
		while ((ivar = [ivars next])) [self outputFileOutMessage:ivar on:aStream];
	}
	
	[aStream printStr:"\treturn (id)self;\n"];
	[aStream printStr:"}\n"];

	return self;
}

- outputFileInCodeOn:aStream
{
	id ivar,ivars,method = implementFileInMethod();

	[method outputFunctionPrototypeOn:aStream];
	[aStream printStr:"\n{\n"];
	
	if ([interface superName]) {
		[self outputSuperFileInMessageOn:aStream];
		ivars = [[interface setOfIdIVars] eachElement];
		while ((ivar = [ivars next])) [self outputFileInMessage:ivar on:aStream];
	 }
	
	[aStream printStr:"\treturn (id)self;\n"];
	[aStream printStr:"}\n"];

	return self;
}

- outputFilerCodeOn:aStream
{
	[aStream printNewline];
	[aStream printNewline];

	if ([options comments]) {
	[aStream printStr:"\n/*\n"];
	[aStream printStr:" *           A S C I I F I L E R   M E T H O D S\n"];
	[aStream printStr:" */\n"];
	[aStream printNewline];
	}

        if ([[interface setOfIdIVars] isEmpty] == NO) {
	    [self outputFileOutCodeOn:aStream];
	    [aStream printNewline];
	    [self outputFileInCodeOn:aStream];
	}

	return self;
}

/******************************************************************************
 *
 * translateEpilogueOn
 *
 *****************************************************************************/

- outputAssignClassGlobalOn:aStream
{
    STR s = [name str];
    [aStream printNewline];

    /* assign struct SHARED to class global */
    /* make this static in -noShared case.  */

    if ([options sharedData] == NO) [aStream printStr:"static "];
    [aStream printStr:"id %s = (id)&_%s;\n",s,s];

    /* in both -noShared and -shared case, output a classGlobalFun() */
    /* (in theory only needed for -noShare) but can do it always */

    [aStream printStr:"id %s *%s(void) { return & %s; }\n",
	[[options bind] str],[self classGlobalFun],s];
	    
    return self;
}

- translateEpilogueOn:aStream
{
	STR s = [name str];

	/* this must be done in the currentImplementation context */
	currentImplementation = self;
	if ([options filerCode]) [self outputFilerCodeOn:aStream];
	currentImplementation = nil;
          
	[self outputDispatchTablesOn:aStream];
	[self outputStructSharedClassOn:aStream];
	[self outputStructSharedMetaClassOn:aStream];
        [self outputAssignClassGlobalOn:aStream];

if ([options comments]) {
	[aStream printNewline];
	[aStream printStr:"\n/*\n"];
	[aStream printStr:" * End of %s implementation\n",s];
	[aStream printStr:" */\n"];
	[aStream printNewline];
}

	[LineTag forceLineTag];
	return self;
}

- makeSetOfIVarsCurrent
{
	[interface makeCurrent];
	return self;
}

- clearCurrentSetOfIVars
{
	[interface clearCurrent];
	return self;
}

- willTranslateMethodImplementationOn:aStream
{
	/* case when we thought we were an interface but at once we find
	 * a method implementation (as opposed to a prototype)
	 * So promote ourselves to an implementation (and interface at
	 * the same time) and translate the prologue before the method impl.
	 */
	 
	 if (isImplementation == NO) {
	 
		/* this is to avoid mistakes with class implementations
		 * without interface declaration part (which we allow,
		 * but do not allow if the super class isn't specified)
		 */
		
		if (interface == self && superName == nil) {
		fatalError("root class without interface not allowed");
		}

		/* another mistake is to put a method impl in an interface */
		if ([self isInterfaceToken]) {
		fatalError("method implementation inside @interface");
		}

	 	[self setIsImplementation:YES];
		[self translateOn:aStream];
	}
	
	return self;
}

- translateOn:aStream
{
	if (isImplementation) {

		if (!interfaceEmitted) [self translateInterfaceOn:aStream];

		if (prologueEmitted) {
			[self translateEpilogueOn:aStream];
		} else {
			[self translatePrologueOn:aStream];
		}
	} else {
		if (!interfaceEmitted) [self translateInterfaceOn:aStream];
	}

	return self;
}

@end

