
/*
 * Portable Object Compiler (c) 1996 by Stes & Lerman  All Rights Reserved.
 */

#include "oclib.h"
#include <string.h> /* strlen */

@implementation LineTag

/*****************************************************************************
 *
 * Amateurish parsing code for line tags.
 * Alternative is to use a mode in lex code, but I decided against that,
 * because it should be feasible to "clean up" the scan.l code by hand-coding
 * here some C for the (simple) case of line directives.
 *
 * Utility function to check whether "aDirective" is something like
 *
 * 	# lineNo "FileName" ... \n
 * or
 * 	# line lineNo "FileName" ... \n
 *
 * If yes, returns by reference the lineNumber and a string object
 * (the filename)
 *
 * NOTE: since HPUX cc is also emitting # lineNo (without fileName)
 * have to support this.
 *
 ****************************************************************************/

static STR GetLineNo(STR s,int *lineNo)
{
	id tmp;
	STR p = s;
	int c,n = 0;

	while ((c = *p++) && '0' <= c && c <= '9') n++;

	tmp = [String chars:s count:n];
	*lineNo = [tmp asInt];
	tmp = [tmp free];

	return p - 1;
}

static BOOL GetFileName(STR s,id *fileName)
{
	STR p = s;
	int c,n = 0;

	while ((c = *p++) && c != '"') n++;

	/* no \" in filenames (yet) (contrary to scan.l:STRING_TOKEN) */
	if (c == '"') {
		*fileName = [String chars:s count:n];return YES;
	} else {
		return NO;
	}
}

static BOOL IsLineDirective(STR aDirective,int *lineNo,id *fileName)
{
	int c;STR p = aDirective;

	/* match the hash sign */
	c = *p++; if (c == '#') { c = *p++; } else { return NO; }

	/* optional whitespace */
	while (c && (c == ' ' || c == '\t')) c = *p++;

	/* accept both "# no name" and "# line no name" */
	if (c == 'l') {
		STR key = "ine"; /* l already consumed */
		if (strncmp(p,key,strlen(key)) == 0) {
			p += strlen(key);c = *p++;
		} else {
			return NO;
		}
	}

	/* optional whitespace */
	while (c && (c == ' ' || c == '\t')) c = *p++;

	/* get the line number */
	if (c && '0' <= c && c <= '9') {
		p = GetLineNo(p - 1,lineNo);c = *p++;
	} else {
		return NO;
	}

	/* optional whitespace */
	while (c && (c == ' ' || c == '\t')) c = *p++;

	/* get the filename */
	if (c && c == '"') {
		return GetFileName(p,fileName);
		/* ignore everything else on the line after the fName */
	} else {
		/* # lineNo style */
		*fileName = nil;
		/* ignore everything else on the line after the fName */
		return YES;
	}
}


/*****************************************************************************
 *
 * Creation
 *
 ****************************************************************************/


static id currentTag;

- next
{
    return next;
}

- setNext:aTag
{
    next = aTag;
    return self;
}

- fileName:(STR)fName lineNo:(int)no directive:s
{
	fileName = fName;
	lineNo = no;
	directive = s;

	/* maintain linked list to be able to traverse ranges of tags */
	/* can't use collection (grows over 32K for large files) */
	if (currentTag) [currentTag setNext:self];

	/* depend on the list nil terminated */
	currentTag = [self setNext:nil];

	return self;
}

+ new
{
	int no;
	STR fName;

	/* we get a +new before yylex() is called */
	if (currentTag) {
		fName = [currentTag fileName];
		no    = [currentTag lineNo] + 1;
	} else {
		id iFile = [options inFile];
		fName = (iFile)?[iFile str]:"stdin";
		no = 1;
	}

	return [[super new] fileName:fName lineNo:no directive:nil];
}

+ str:(STR)aDirective
{
	int no;
	id s,d;
	STR fName;
	id current = currentTag;

	if (IsLineDirective(aDirective,&no,&s)) {
		if (s) {
			fName = [s strCopy];s = [s free];
		} else {
			/* "# lineNo" style */
			fName = [current fileName];
			if (!fName) warnError("# lineNo without filename");
		}

		d = nil;
		return [[super new] fileName:fName lineNo:no directive:d];
	} else {

#if 0
		/* cc pragma's, indents etc. */
		fName = [current fileName];
		no    = [current lineNo]; /* same number as preceding ! */
		d     = [String str:aDirective];
		self = [[super new] fileName:fName lineNo:no directive:d];
	
		/* add additional node for newline in directive */
		return [LineTag new];
#endif

		/* cc pragma's, indents etc. */
		fName = [current fileName];
		no    = [current lineNo] + 1;
		d     = [String str:aDirective];
		return [[super new] fileName:fName lineNo:no directive:d];
	}
}

+ currentTag
{
	return currentTag;
}

- (int) lineNo
{
	return lineNo;
}

- (STR) fileName
{
	return fileName;
}

+ (int) currentLineNo
{
	return [currentTag lineNo];
}

+ (STR) currentFileName
{
	return [currentTag fileName];
}

static id outCurrent;
static BOOL forceOutput;

static void OutputLineTagOn(int no,STR fName,id aStream)
{
	id format = [options lineTagFormat];
	[aStream printStr:[format str],no,fName];
	[aStream printNewline];
}

+ outputOn:aStream
{
	STR fName;

	if ([options srcFileName]) {
		fName = [[options srcFileName] str];
	} else {
		fName = "stdin";
	}

	if ([options lineTags]) OutputLineTagOn(1,fName,aStream);
        outCurrent = [[super new] fileName:fName lineNo:1 directive:nil];

	return self;
}

+ forceLineTag
{
	forceOutput = YES;return self;
}
	
- outputDirectiveOn:aStream
{
	/* the pragma includes a newline */
	if (directive) {
		/* [self outputOn:aStream]; */
		[aStream printObject:directive];
	}
	return self;
}

- outputOn:aStream
{
	if ([options lineTags]) OutputLineTagOn(lineNo,fileName,aStream);
	return self;
}

+ (BOOL) areSuccessors:a:b
{
	if ([a lineNo] + 1 == [b lineNo]) {
		return strcmp([a fileName],[b fileName]) == 0;
	} else {
		return NO;
	}
}

- (BOOL) comesBefore:b
{
    id nextTag = next;

    while (nextTag) {
	if (nextTag == b) return YES;
	nextTag = [nextTag next];
    }

    return NO;
}

+ outputOn:aStream until:offset
{
	/* note that we cannot assert that offset > outCurrent */
	/* indeed, it's possible to dump a piece of code (like */
	/* in a struct for @defs) with linetags that are lower */
	/* than outCurrent (ivars for the @defs being declared */
	/* in some header file, long before the @defs usage)   */

	/* just output whitespace if there was no newline */
	if (forceOutput==NO && outCurrent==offset) {
		[aStream printChar:' '];return self;
	}

	[aStream printNewline];

	/* a single newline is sufficient here if they're successors */
	if (forceOutput || [self areSuccessors:outCurrent:offset]==NO) {

	        if ([outCurrent comesBefore:offset]) {
		     id aTag = outCurrent;

		     /* only do pragma's etc. - no consecutive newlines */
		     while (aTag != offset) {
		        aTag = [aTag next];
			if (aTag) [aTag outputDirectiveOn:aStream];
		     }
		}

		/* just one lineTag */
		[offset outputOn:aStream];
	}

	forceOutput = NO;
	outCurrent  = offset;

	return self;
}

+ finishOutputOn:aStream
{
	id aTag = outCurrent;

	[aStream printNewline];

	do {
		aTag = [aTag next];
		if (aTag) [aTag outputDirectiveOn:aStream];
	} while (aTag);

	[aStream printNewline];
	return self;
}

@end

