
/*
 * Portable Object Compiler (c) 1996 by Stes & Lerman  All Rights Reserved.
 */

#include "oclib.h"
#include "assert.h"

@implementation MethodInfo

static id methodPrototypes;

- free
{
    return nil;
}

- prototype
{
	id aName;

	if (!methodPrototypes) methodPrototypes = [Dictionary new];

	aName = [self fullMethodName];

	if ([methodPrototypes atKey:aName]) {
		/* should check for types here */
	} else {
		[methodPrototypes atKey:aName put:self];
	}

	return self;
}

- implement
{
	id aClass = [ClassDefinition current];
	
	/* implementing implicitely prototypes */
	/* also - prototype will check types   */

	[self prototype];

	/* add to the class implementation */

	if (aClass) {
		if ([self isFactory]) {
			[aClass addClsDispatchTable:self];
		} else {
			[aClass addNstDispatchTable:self];
		}
	} else {
		fatalError("method implementation outside class definition");
	}

	return self;
}

+ findPrototypeFor:aName
{
	return [methodPrototypes atKey:aName];
}

static void warnNotFound(id aName)
{
	if ([options warnUndefinedMethod]) {
		char b[8192];
		sprintf(b,"argument types for method %s default to 'id'",[aName str]);
		warnError(b);
	}
}

- argumentsDefaultToId
{
	[self setResultType:nil];
	/* arguments have been defined as "id" too */
	assert([arguments size] == [argumentTypes size]);
	return self;
}

+ findOrWarnPrototype:aMethod
{
	id m;
	id aName = [aMethod fullMethodName];
	
	if (!(m = [self findPrototypeFor:aName])) {
		warnNotFound(aName);
		[aMethod argumentsDefaultToId];
		[aMethod prototype];
		m = [self findPrototypeFor:aName];
		assert(m != nil);
	}
	return m;
}


- (BOOL) isEqual:aMethod
{
	return [name isEqual:[aMethod fullMethodName]];
}


- fullMethodName
{
	return name;
}

- functionNameForClass:aClassName
{
	char *q;
	id funcName,aBuffer = [name copy];
	
	/* convert colons to underscores */
	for(q=[aBuffer str];*q!=0; q++){ if(*q==':')*q='_'; }

	/* 'c_' prefix denotes class/factory method
	 * 'i_' stands for instance method
	 * Include classname so that we can have multiple methods implementations
	 * (for different classes) in same file.
	 */
	
	if([self isFactory]) {
	funcName = [String sprintf:"c_%s_%s",[aClassName str],[aBuffer str]];
	} else {
	funcName = [String sprintf:"i_%s_%s",[aClassName str],[aBuffer str]];
	}

	return funcName;
}

- setIsFactory:(BOOL)flag
{
	isFactory = flag;return self;
}

- (BOOL) isFactory
{
	return isFactory;
}

- setResultType:aNode
{
	resultType = aNode;return self;
}


- addNameComponent:aString
{
	if (!name) name = [String new];
	[name concatSTR:[aString str]];
	return self;
}

- addColonToName
{
	assert(name);
	[name concatSTR:":"];
	return self;
}

- addEllipsis:aToken
{
	ellipsis = aToken;
	return self;
}

- addArgumentType:aNode
{
	if (!argumentTypes) argumentTypes = [OrdCltn new];
	[argumentTypes add:aNode];
	/* protect ourselves against missing |processCurrent| calls */
	assert([argumentTypes size] < 100);
	return self;
}

- addIdArgumentType
{
	return [self addArgumentType:[Token newToken:"id"]];
}

- addArgument:aNode
{
	if (!arguments) arguments = [OrdCltn new];
	[arguments add:aNode];
	/* protect ourselves against missing |processCurrent| calls */
	assert([arguments size] < 100);
	return self;
}

static void warnWithMethod(STR name,int expected,int found)
{
    char b[8192];
    sprintf(b,"method %s%d followed by %d arguments",name,expected,found);
    warnError(b);
}

- checkForMethodWith
{
    id firstEl;
    int expected,n;
    static id methodsLikeWith;

    if (!methodsLikeWith) {
	methodsLikeWith = [Set add:[String str:"with:"]];
    }

    /* stes 1/98
     * suggestion Marc Boschma marcb@c3.telstra-mm.net.au
     * to count number of arguments for +with: method
     * we could also automatically generate the count
     * so that the user doesn't have to tell us the number
     * of arguments, but that would be non-portable
     * across Objective-C compilers
     */

    if (![methodsLikeWith contains:[self fullMethodName]]) return self; 

    /* way we deal with comma-list arguments is kind-o-fishy since
     * it counts here as one colon argument +with:oneArg where
     * oneArg is actually in itself a Node (expression) with the
     * commalist.  we want the firstElement of the list
     */

    assert([arguments size] == 1);
    firstEl = [[arguments lastElement] subnodes];
    n = [firstEl size];assert(n >= 1);n=(n-1)/2; /* commas are in there ...*/

    expected = [[firstEl firstElement] asInt]; /* gives 0 if symbolic */

    /* it might be symbolic (expected=0), then we cant check */
    if (expected && (n != expected)) {
	warnWithMethod([[self fullMethodName] str],expected,n);
    }

    return self;
}


static id stdMsgArgs,stdMsgTypes;

- stdMsgArgs
{
    if (stdMsgArgs) return stdMsgArgs;
    stdMsgArgs=[OrdCltn new];
    [stdMsgArgs add:[Identifier str:"self"]];
    [stdMsgArgs add:[Identifier str:"_cmd"]];
    return stdMsgArgs;
}

- stdMsgTypes
{
    id aClass,s;
    if (stdMsgTypes) return stdMsgTypes;
    stdMsgTypes=[OrdCltn new];

    /* 11/97 - changed id type to the current class type */
    aClass = [ClassDefinition currentName];
    s = [String sprintf:"struct %s_PRIVATE *",[aClass str]];
    [stdMsgTypes add:[Identifier string:s]];
    [stdMsgTypes add:[Identifier str:"SEL"]];
    return stdMsgTypes;
}

- addLocalsTo:aCompound
{
    [aCompound addLocalVars:[self stdMsgArgs] types:[self stdMsgTypes]];

    if (arguments) {
        [aCompound addLocalVars:arguments types:argumentTypes];
    }

    return self;
}


- outputResultTypeOn:aStream
{
	/* return type */
	if (resultType) {
		[resultType outputWithoutTagsOn:aStream];
	} else {
		[aStream printStr:"id"];
	}
	return self;
}

- outputCastOn:aStream
{
	[self outputResultTypeOn:aStream];

	/* function pointer */
	[aStream printStr:" (*) "];

	/* argument types */
	[aStream printStr:"("];
	[self outputArgumentTypesOn:aStream];
	[aStream printStr:")"];

	return self;
}

- outputArgumentTypesOn:aStream
{
	int i,n;

	n = (argumentTypes)?[argumentTypes size]:0;

	/* first the "self" and "_cmd" types */
	[aStream printStr:"id,SEL"];
	
	for (i=0;i<n;i++) {
		[aStream printStr:","];
		[[argumentTypes at:i] outputWithoutTagsOn:aStream];
	}

	if (ellipsis) {
		[aStream printStr:","];
		[ellipsis outputWithoutTagsOn:aStream];
	}

	return self;
}

- outputArgumentProtosOn:aStream
{
	int i,m,n;
	id aClass;

	m = (arguments)?[arguments size]:0;
	n = (argumentTypes)?[argumentTypes size]:0;

	if (m!=n) {
	    warnError("internal error: number arguments != number types");
	}

	/* first the "self" and "_cmd" types 
	 *
	 * stes - feb 3, 1997
	 * because we changed for Microsoft C the def of "id"
	 * as emitted by objc (see Module.m)
	 * we need to introduce a cast here.
         * This isn't bad actually, since, when we will do more than 1
	 * object per file, this would be the place to the cast.
	 *
	 * stes - apr29, 1997
	 * Prepended name of class to _PRIVATE.
	 *
	 * 11/97 our philosophy is that self just an 'id' variable,
	 * but many people ask for instance variable access in a debugger
	 * so we now type this here as a struct PRIVATE
	 * (as opposed to how it was before, where the cast was done
	 * only at the time of variable translation)
	 */

	aClass = [ClassDefinition currentName];
	[aStream printStr:"struct %s_PRIVATE *self,SEL _cmd",[aClass str]];
	
	for (i=0;i<n;i++) {
		[aStream printStr:","];
		[[argumentTypes at:i] outputWithoutTagsOn:aStream];
		[[arguments at:i] outputWithoutTagsOn:aStream];
	}

	if (ellipsis) {
		[aStream printStr:","];
		[ellipsis outputWithoutTagsOn:aStream];
	}

	return self;
}

- translateArgumentsOn:aStream
{
	int i,n;
	
	n = (arguments)?[arguments size]:0;

	/* always preceded by self and SEL so output a comma */
	
	for (i=0;i<n;i++) {
		[aStream printStr:","];
		[[arguments at:i] translateOn:aStream];
	}

	return self;
}

- outputFunctionPrototypeOn:aStream
{
	id s;

	[aStream printStr:"static "];
	[self outputResultTypeOn:aStream];
	s = [self functionNameForClass:[ClassDefinition currentName]];
	[aStream printStr:" %s(",[s str]];
	[self outputArgumentProtosOn:aStream];
	[aStream printStr:")"];

	return self;
}

@end

