
/*
 * Portable Object Compiler (c) 1996 by Stes & Lerman  All Rights Reserved.
 */

#include "oclib.h"

@implementation MessageExpression

static int varCounter = 0;
static int inlineCacheCounter = 0;

- processReceiver:aReceiver
{
	receiver = aReceiver;

	/* count of the temporary variable that will hold the receiver */
	tmpVar   = [String sprintf:"objcT%i",varCounter++];

	/* name of the inlineCache (if we're doing this) */
	if ([options inlineCache]) {
	    inlineCache = [String sprintf:"objcIC%i",inlineCacheCounter++];
	}

	isReceiverSuper = [receiver isEqualSTR:"super"];

	if (isReceiverSuper) {
	    id cName = [ClassDefinition currentName];
	    if (!cName) fatalError("message to super outside @implementation");
	}

	/* idea here is that linetag associated to the entire msgxpr
	 * is equal to linetag of left '['; this tag must be emitted before
	 * any generated code
	 */

        linetag = [[subnodes at:0] lineTag];

	return self;
}

- processMessageArguments:aMethodDeclarator
{
	methodMsg   = [aMethodDeclarator method];
	methodName  = [methodMsg fullMethodName];
	methodProto = [MethodInfo findOrWarnPrototype:methodMsg];
	
	[methodMsg checkForMethodWith]; /* +with: gadget */

	return self;
}

- addTempVarsTo:aStmt
{
	/* add this tmp var to the list to be declared by the statement */
	[aStmt addTempVar:tmpVar];

	/* handle subnodes (such as the receiver) */
	return [super addTempVarsTo:aStmt];
}

- addInlineCachesTo:aStmt
{
	assert([options inlineCache]);

	/* add this tmp var to the list to be declared by the statement */
	[aStmt addInlineCache:inlineCache];

	/* handle subnodes (such as the receiver) */
	return [super addInlineCachesTo:aStmt];
}


- translateOn:aStream
{
	[aStream printNewline];
        [linetag outputOn:aStream];

	if (isReceiverSuper) {
		[self translateSuperMessageOn:aStream];
	} else {
		[self translateMessageOn:aStream];
	}

	return self;
}

- genAssignReceiverToTempOn:aStream
{
	STR tmpVarName = [tmpVar str];
	[aStream printStr:"%s = ",tmpVarName];
	[LineTag forceLineTag];
	[receiver translateOn:aStream];
	return self;
}

- genAssignSuperToTempOn:aStream
{
    STR tmpVarName = [tmpVar str];
    id cName = [ClassDefinition currentName];
    if (!cName) return self;

    /* the meaning of "super" depends on whether we're in a factory */
    /* or instance method implementation                          */

    if ([currentMethodImplementation isFactory]) {
    /* we need a cast from struct _SHARED to id */
	if ([options sharedData]) {
	[aStream printStr:"%s = (id)(__%s.clsSuper)",tmpVarName,[cName str]];
	} else {
	STR x = [[ClassDefinition current] metaStructFun];
	[aStream printStr:"%s = (id)(%s()->clsSuper)",tmpVarName,x];
	}
    } else {
	if ([options sharedData]) {
	[aStream printStr:"%s = (id)(_%s.clsSuper)",tmpVarName,[cName str]];
	} else {
	STR x = [[ClassDefinition current] classStructFun];
	[aStream printStr:"%s = (id)(%s()->clsSuper)",tmpVarName,x];
	}
    }

    return self;
}

- genImpCallOn:aStream inSuper:(BOOL)inSuper
{
	STR tmp = [tmpVar str];
	STR funcName = (inSuper)?"_impSuper":"_imp";

	/* regular, non-optimized messenger call */

	[aStream printStr:"%s(%s,",funcName,tmp];
	[aStream printSelector:methodName];
	[aStream printStr:")"];

	return self;
}

- genInlineImpCallOn:aStream inSuper:(BOOL)inSuper
{
	STR getclass;
	STR tmp = [tmpVar str];
	STR cache = [inlineCache str];

	/* this is the optimized messenger call */

	/* check for messages to nil */
	[aStream printStr:"(%s)?(",tmp];

	/* check for class pointer equal */
	[aStream printStr:"("];
	getclass = (inSuper)?"%s":"%s->isa";
	[aStream printStr:getclass,tmp];
	[aStream printStr:" == %s.cls)?%s.imp:",cache,cache];

	[aStream printStr:"("];
 	/* cache failure, do regular lookup */
	[aStream printStr:"%s.cls=",cache];
	[aStream printStr:getclass,tmp];
	[aStream printStr:",%s.imp=",cache];
	[self genImpCallOn:aStream inSuper:inSuper];
	[aStream printStr:")"];

	[aStream printStr:"):_nilHandler"];
	return self;
}

- genCallOn:aStream inSuper:(BOOL)inSuper
{
	/* call messenger, cast return value, and dereference */

	[aStream printStr:"(*("];
	[methodProto outputCastOn:aStream];
	[aStream printStr:")"];
	
	if ([options inlineCache]) {
	    [aStream printStr:"("];
	    [self genInlineImpCallOn:aStream inSuper:inSuper];
	    [aStream printStr:")"];
	} else {
	    [self genImpCallOn:aStream inSuper:inSuper];
	}
	
	[aStream printStr:")"];

	return self;
}

- translateMessageArgumentsOn:aStream inSuper:(BOOL)inSuper
{
	/* temp var for receiver & selector _cmd */

	if (inSuper) {
	    /* can be inside a block ! then we output scope->self */
	    [Identifier translateSelfAsIdOn:aStream];
	    [aStream printStr:","];
	} else {
	    STR tmp = [tmpVar str];
	    [aStream printStr:"%s,",tmp];
	}

	[aStream printSelector:methodName];

	[LineTag forceLineTag];
	[methodMsg translateArgumentsOn:aStream];/* begins with comma */

	return self;
}

- genMessageOn:aStream inSuper:(BOOL)inSuper
{
	[self genCallOn:aStream inSuper:inSuper];
	[aStream printStr:"("];
	[self translateMessageArgumentsOn:aStream inSuper:inSuper];
	[aStream printStr:")"];

	return self;
}

- translateMessageOn:aStream
{
	[aStream printStr:"("];
	[self genAssignReceiverToTempOn:aStream];
	[aStream printStr:","];
	[self genMessageOn:aStream inSuper:NO];
	[aStream printStr:")"];

	[LineTag forceLineTag];return self;
}

- translateSuperMessageOn:aStream
{
	[aStream printStr:"("];
	[self genAssignSuperToTempOn:aStream];
	[aStream printStr:","];
	[self genMessageOn:aStream inSuper:YES];
	[aStream printStr:")"];

	[LineTag forceLineTag];return self;
}

@end

